"""Solar position calculates the sun zenith and azimuth angle.

Implemented according to Blanco-Muriel et al., Solar Energy 70(5), 2001.

This module contains the following functions:
date2jd - compute the time conversion to Julian days
eclipse - compute the solar eclipse
ra - compute the right ascension
delta - compute the declination angle
lmst - compute the local time for the calculation of the hour angle
zenith - compute the zenith angle
parallax - compute the parallax
azimuth - compute the azimuth angle
calc_zenith - compute the zenith and azimuth angle in degrees
"""

import datetime as dt

import numpy as np


def date2jd(dates):
    """Compute the time conversion to Julian days.

    :param dates: a datetime index

    :return: An array with the converted times
    """
    y = dates.year.to_numpy()
    m = dates.month.to_numpy()
    d = dates.day.to_numpy()
    decimal_hours = dates.hour.to_numpy() + (dates.minute.to_numpy()
                                             + dates.second.to_numpy()/60.)/60.
    aux1 = ((m-14.)/12.).astype(int)
    aux2 = ((1461.*(y+4800.+aux1))/4. + (367.*(m-2.-12.*aux1))/12.
            - (3.*((y+4900.+aux1)/100.))/4. + d - 32075.).astype(int)
    return aux2 - 0.5 + decimal_hours/24.


def eclipse(n):
    """Compute the solar eclipse, i.e. ecliptic longitude and obliquity.

    :param n: Julian days since 2000-01-01 12:00 UTC

    :return: The ecliptic longitude and obliquity of ecliptic
    """
    omega = 2.14292 - 0.0010394594*n
    L = 4.8950630 + 0.017202791698*n
    g = 6.2400600 + 0.0172019699*n
    l = (L + 0.03341607*np.sin(g) + 0.00034894*np.sin(2.*g) - 0.0001134
         - 0.0000203*np.sin(omega))
    ep = 0.4090928 - 6.2140e-9*n + 0.0000396*np.cos(omega)
    return l, ep


def ra(l, ep):
    """Compute the right ascension.

    Conversion from ecliptic to celestial coordinates.

    :param l: the ecliptic longitude
    :param ep: the obliquity of ecliptic

    :return: The right ascension
    """
    res = np.arctan2(np.cos(ep)*np.sin(l), np.cos(l))
    res[res < 0.] += 2.*np.pi  # value must be between 0 and 2pi
    return res


def delta(l, ep):
    """Compute the declination angle.

    Conversion from ecliptic to celestial coordinates.

    :param l: the ecliptic longitude
    :param ep: the obliquity of ecliptic

    :return: The declination angle
    """
    return np.arcsin(np.sin(ep)*np.sin(l))


def lmst(n, hour, lon):
    """Compute the local time for the calculation of the hour angle.

    :param n: Julian days since 2000-01-01 12:00 UTC
    :param hour: the hours taken from each given timestamp
    :param lon: the station's longitude

    :return: The local time
    """
    gmst = 6.6974243242 + 0.0657098283*n + hour
    return (gmst*15. + lon) * np.pi/180.


def zenith(lat, h_angle, decl):
    """Compute the zenith angle.

    :param lat: the station's latitude
    :param h_angle: the hour angle
    :param decl: the declination angle

    :return: The zenith angle
    """
    rlat = lat * np.pi/180.
    return np.arccos(np.cos(rlat)*np.cos(h_angle)*np.cos(decl)
                     + np.sin(decl)*np.sin(rlat))


def parallax(z):
    """Compute the parallax.

    :param z: the zenith angle

    :return: The parallax
    """
    EMR = 6371.01  # mean earth radius in km
    AU = 149597890.  # astronomical unit
    return EMR/AU * np.sin(z)


def azimuth(lat, h_angle, decl):
    """Compute the azimuth angle.

    :param lat: the station's latitude
    :param h_angle: the hour angle
    :param decl: the declination angle

    :return: The azimuth angle
    """
    rlat = lat * np.pi/180.
    res = np.arctan2(-np.sin(h_angle), (np.tan(decl)*np.cos(rlat)
                                        - np.sin(rlat)*np.cos(h_angle)))
    res[res < 0.] += 2.*np.pi  # value must be between 0 and 2pi
    return res


def calc_zenith(dates, lat, lon=0., local_time=True):
    """Compute the zenith and azimuth angle in degrees.

    :param dates: a datetime index
    :param lat: the station's latitude
    :param lon: the station's longitude
    :param local_time: a boolean denoting whether the times are in local
                       time or UTC

    :return: The zenith and azimuth angle in degrees
    """
    utc_dates = dates
    if local_time:
        utc_dates = dates - dt.timedelta(hours=lon/15.)
    jd = date2jd(utc_dates)
    n = jd - 2451545.0  # elapsed Julian days since 2000-01-01 12:00 UTC
    l, ep = eclipse(n)
    rasc = ra(l, ep)
    d = delta(l, ep)
    lt = lmst(n, utc_dates.hour.to_numpy(), lon)
    h_angle = lt - rasc
    z = zenith(lat, h_angle, d)
    px = parallax(z)
    az = azimuth(lat, h_angle, d)
    return (z+px) * 180./np.pi, az * 180./np.pi
